#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <errno.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>
#include <sys/time.h>

#include <libcc32.h> /* the header of the shared library */
#include "daq.h"

#define DEVICE_NAME "/dev/cc32_1"

CC32_HANDLE handle;
 
#define CSET_I cc32_read_word(handle, 27, 0, 0)
#define CREM_I cc32_read_word(handle, 27, 1, 0)

/* definiram lokacije enot*/
#define NTDC 19         /* TDC LeCroy 2277*/
#define NADC 15         /* ADC AD811*/

int ctrlc=0;


#define TRUE 1
#define FALSE 0


int timer_out;
struct sigaction oact;
void timerast (int signumber)
{
  timer_out = TRUE;
  fprintf(stderr,"TIMEOUT !!!\n");
}

void tmlnk (int tout)
{
  timer_out = FALSE;
  struct sigaction act;
  struct itimerval tdelay;

  act.sa_handler = timerast;
  sigemptyset (&act.sa_mask);
  act.sa_flags = 0;
  tdelay.it_value.tv_sec = tout / 100;
  tdelay.it_value.tv_usec = 10000 * (tout % 100);
  tdelay.it_interval.tv_sec = 0;
  tdelay.it_interval.tv_usec = 0;

  if (sigaction (SIGALRM, &act, &oact) < 0)
    {
      perror ("sigaction(tmlnk)");
      exit (EXIT_FAILURE);
    }
  if (setitimer (ITIMER_REAL, &tdelay, NULL) < 0)
    {
      perror ("setitimer(tmlnk)");
      exit (EXIT_FAILURE);
    }
}

void tmulk ()
{
  struct itimerval tdelay;

  tdelay.it_value.tv_sec = 0;
  tdelay.it_value.tv_usec = 0;
  tdelay.it_interval.tv_sec = 0;
  tdelay.it_interval.tv_usec = 0;

  if (setitimer (ITIMER_REAL, &tdelay, NULL) < 0)
    {
      perror ("setitimer(tmulk)");
      exit (EXIT_FAILURE);
    }
  if (sigaction (SIGALRM, &oact, NULL) < 0)
    {
      perror ("sigaction(tmulk)");
      exit (EXIT_FAILURE);
    }
}



void CAMAC_status(void)
{
  unsigned short id,fpga,modid,q,x,inhibit,lam;
  unsigned short dum;
  
  dum = cc32_read_word(handle, 0, 0, 0);

  id   = (dum & 0xF000) >> 12;
  fpga = (dum & 0x0F00) >> 8;   
  modid =  (dum & 0x00F0) >> 4; 
  q      =  (dum & 0x0008) >> 3;
  x      =  (dum & 0x0004) >> 2;
  inhibit=  (dum & 0x0002) >> 1;
  lam    =  (dum & 0x0001);	
  printf ("CC32 CSR ID=%x FPGA=%x modID=%x Q=%d X=%d INHIBIT=%d LAM=%d \n",id,fpga,modid, q,x,inhibit,lam);
}

int daq::init(){
      // write a reset to CC32
      cc32_write_word( handle, 31, 0, 16, 0x55aa);
      // enable all LAM channels
      cc32_write_long(handle, 28, 1, 16, 0xFFFFFFFF);
      //CCCZ;
      cc32_read_long_all(handle, 0, 1, 0); 
      //CCCC;
      cc32_read_long_all(handle, 0, 0, 0);     
      
      // set status register
      // int tdcreg2277=0x8C00; // status register common start (green LED),Front panel timeout ,leading+trailing
      //  tdcreg2277= 0xC00;   //  status register common stop (red LED)                        ,leading + trailing edge recording
      int  tdcreg2277= 0xCC00; //  status register common start (green LED),8192 ns timeout     ,leading+trailing

      cc32_write_word(handle,NTDC, 0, 17, tdcreg2277);
      int dum = cc32_read_word(handle,NTDC, 0, 17);
      printf("SetRegister 0x%04x GetRegister 0x%04x\n",tdcreg2277,dum);

      CAMAC_status();
      return 0;
}

int daq::connect(){
    char *fname = DEVICE_NAME;
    int error;
    if ((error = cc32_open(fname, &handle))) {
        fprintf(stderr, "%s: %s\n", fname, strerror(error));
        exit(1);
    } else printf("cc32 opened\n");
    return 0;
}
  
int daq::disconnect(){ 
  /* zakljuci */
   cc32_close(handle);
   printf("cc32 closed.\n");
  return 0;
}


int daq::event(unsigned int *data, int maxn, int *ierr){
   unsigned int dum=1;
   int q,x;


   *ierr=0;
   cc32_read_word_qx(handle, NTDC, 0, 9, &q, &x);   // clear TDC 
   cc32_read_word_qx(handle, NTDC, 0, 10, &q, &x);  // clear LAM 
   cc32_write_word(handle, NTDC, 0, 26, dum);       // enable LAM

   cc32_read_word_qx(handle, NADC, 0, 9, &q, &x);   // clear ADC 2249A
   cc32_read_word_qx(handle, NADC, 0, 10, &q, &x);  // clear LAM 2249A
   cc32_write_word(handle, NADC, 0, 26, dum);       // enable ADC LAM 2249A



   //   cc32_read_word_qx(handle, NADC, 12, 11, &q, &x);  // clear all registers AD811 
   //   cc32_write_word(handle, NADC, 12, 26, dum);       // enable LAM AD811

   // test LAM
   int tout=200; /* 1/100 of a second */
   tmlnk (tout);
   do {
      if (ctrlc) break;
      cc32_read_long_qx(handle, NADC, 0, 8, &q, &x);
   } while (q==0 && fStop==0 && !timer_out); 
   tmulk();
   if (timer_out ) *ierr=1;
   //printf("q=%d fStop=%d\n", q, fStop);  
   // read data
   int count=0;

   //AD811

   const int ndata=8;

   // test LAM for AD811
   // do{  
   //  if (ctrlc) break; 
   //  cc32_read_long_qx(handle, NADC, 12, 8, &q, &x);
   //} while (q==0 && fStop==0);

   for ( int j=0 ; j<ndata ; j++ ) 
     //     data[count++] = cc32_read_long_qx(handle, NADC, j, 2, &q, &x); //AD811
     data[count++] = cc32_read_long_qx(handle, NADC, j, 0, &q, &x); //2249A


   while(1) {
     if (fStop) break;
     if (ctrlc) break;
     dum = cc32_read_long_qx(handle, NTDC, 0, 0, &q, &x); 
     //printf("count=%d q=%d x=%d 0x%x\n",count,q,x, dum);
     if (q == 0) {
       // test buffering in progress
       //cc32_read_word_qx(handle, NTDC, 0, 27, &q, &x); // a to spploh gre ker je 27 (write)
       cc32_write_word(handle, NTDC, 0, 27,dum);
       dum = cc32_read_word(handle, 0, 0, 0);
       q= (dum & 0x0008) >> 3;
       // namesto zgornjega klica
       if (q == 1) continue;
       break;
     }
     if (count<maxn) data[count] = dum;    
     count++;        
   }

   return count*sizeof(int);
}



daq::daq(){
fStop=0;
connect();
init();
}

daq::~daq(){
disconnect();
}

#ifdef MAIN
void CatchSig (int signumber)
{
   ctrlc = 1;
}


int main (int argc, char **argv){
  
  int neve=1000000;
  char *fname="test.dat";
  if (argc==1) {
    printf("Uporaba: %s stevilo_dogodkov ime_datoteke\n",argv[0]);
    printf("Meritev prekini s Ctrl-C, ce je nabranih dogodkov ze dovolj\n");
    exit(0);
  } 
  if (argc>1) neve  = atoi(argv[1]);
  if (argc>2) fname = argv[2];
  
  
  // intercept routine
  if (signal (SIGINT, CatchSig) == SIG_ERR) perror ("sigignore");
   

  // odpremo datoteko za pisanje
  FILE *fp=fopen(fname,"w");
  
  int hdr[4]={1};
  
  #define BSIZE 10000
  unsigned int data[BSIZE];
  daq *d= new daq();
  d->init();
  int ierr=0;
  for (int ieve=0;ieve!=neve && !ctrlc ;ieve++){
    int nb=d->event(data,BSIZE, &ierr);
    if (ierr){
      ieve--;
      continue;
    }
    // zapis v datoteko   
    hdr[1]=nb+4*sizeof(int);
    hdr[2]=time(NULL);
    hdr[3]=ieve;

    fwrite(hdr,   sizeof(int),4 , fp); 
    fwrite(data,   sizeof(int),nb, fp);
    // napolni histograme
    if (ieve%1000!=0) continue;
    int nint=nb/sizeof(int);
    for (int i=0;i<8;i++){
      unsigned short adc =   data[i]&0xFFFF;	
      printf("nev=%4d %d. ADC data=%d\n",ieve,i,adc);
    }
    for (int i=8;i<nint;i++){
       unsigned short tdc =   data[i]&0xFFFF;		       
       unsigned short ch  =  (data[i] >> 17 ) &0x1F;               	  	
       unsigned short edge = (data[i] >> 16 ) &0x1;
       printf("nev=%4d %d. [ch=%2d] egde=%d data=%d\n",ieve,i,ch,edge,tdc);
    }
    
  } 
  fclose(fp);
  printf("Podatki so v datoteki %s\n", fname);
  delete d;

  return 0;
}
#endif
